import json
import pytest

from src.lunar_policy import Check, Node
from src.lunar_policy.result import NoDataError


class TestCheckNoData:
    def test_can_report_no_data(self, capsys):
        node = Node.from_component_json({})

        with Check('test', node=node) as c:
            c.assert_true(c.get_value('.not.a.path'))

        captured = capsys.readouterr()
        results = json.loads(captured.out)['assertions']
        assert len(results) == 1
        assert results[0]['result'] == 'no-data'
        assert '.not.a.path' in results[0]['failure_message']
        assert results[0]['op'] == 'fail'

    def test_can_report_no_data_with_get(self, capsys):
        node = Node.from_component_json({})

        with Check('test', node=node) as c:
            c.get_value('.not.a.path')

        captured = capsys.readouterr()
        results = json.loads(captured.out)['assertions']
        assert len(results) == 1
        assert results[0]['result'] == 'no-data'
        assert '.not.a.path' in results[0]['failure_message']
        assert results[0]['op'] == 'fail'

    def test_can_report_no_data_with_get_all(self, capsys):
        node = Node.from_component_json({})

        with Check('test', node=node) as c:
            c.get_all_values('.not.a.path')

        captured = capsys.readouterr()
        results = json.loads(captured.out)['assertions']
        assert len(results) == 1
        assert results[0]['result'] == 'no-data'
        assert '.not.a.path' in results[0]['failure_message']
        assert results[0]['op'] == 'fail'

    def test_assert_none_is_no_data_workflows_not_finished(self, capsys):
        node = Node.from_component_json({})

        with Check('test', node=node) as c:
            c.assert_true(None)

        captured = capsys.readouterr()
        results = json.loads(captured.out)['assertions']
        assert len(results) == 1
        assert results[0]['result'] == 'no-data'
        assert 'None' in results[0]['failure_message']
        assert results[0]['op'] == 'fail'

    def test_assert_none_is_no_data_workflows_finished(self, capsys):
        node = Node.from_component_json({}, bundle_info={'workflows_finished': True})

        with pytest.raises(NoDataError):
            with Check('test', node=node) as c:
                c.assert_true(None)

    def test_exit_check_early_on_no_data(self, capsys):
        node = Node.from_component_json(
            {
                'value': True,
            }
        )

        with Check('test', node=node) as c:
            c.assert_true(c.get_value('.value'))
            c.assert_false(c.get_value('.not.a.path'))
            c.assert_equals('should not run', 'should not run')

        captured = capsys.readouterr()
        result = json.loads(captured.out)
        assert len(result['assertions']) == 2
        assert all(a['op'] != 'equals' for a in result['assertions'])

    def test_exit_check_early_on_no_data_get(self, capsys):
        node = Node.from_component_json({})

        with Check('test', node=node) as c:
            c.get_value('.not.a.path')
            c.assert_equals('should not run', 'should not run')

        captured = capsys.readouterr()
        result = json.loads(captured.out)
        assert len(result['assertions']) == 1

    def test_exit_check_error_on_no_data_workflows_not_finished(self, capsys):
        node = Node.from_component_json({}, bundle_info={'workflows_finished': False})

        with Check('test', node=node) as c:
            c.get_value('.not.a.path')
            c.assert_equals('should not run', 'should not run')

        captured = capsys.readouterr()
        result = json.loads(captured.out)
        assert len(result['assertions']) == 1
        assert result['assertions'][0]['result'] == 'no-data'
        assert '.not.a.path' in result['assertions'][0]['failure_message']

    def test_exit_check_error_on_no_data_workflows_finished(self, capsys):
        node = Node.from_component_json({}, bundle_info={'workflows_finished': True})

        with pytest.raises(ValueError):
            with Check('test', node=node) as c:
                c.assert_true(c.get_value('.not.a.path'))
                c.assert_equals('should not run', 'should not run')

        captured = capsys.readouterr()
        result = json.loads(captured.out)
        assert len(result['assertions']) == 1
        assert result['assertions'][0]['result'] == 'error'
        assert result['assertions'][0]['op'] == 'fail'
        assert '.not.a.path' in result['assertions'][0]['failure_message']

    def test_exit_check_error_on_manual_no_data_workflows_not_finished(self):
        node = Node.from_component_json({}, bundle_info={'workflows_finished': False})

        with Check('test', node=node):
            raise NoDataError('No data found')

    def test_exit_check_error_on_manual_no_data_workflows_finished(self):
        node = Node.from_component_json({}, bundle_info={'workflows_finished': True})

        with pytest.raises(NoDataError):
            with Check('test', node=node):
                raise NoDataError('No data found')

    def test_suppress_on_no_data(self, capsys):
        node = Node.from_component_json({})
        with Check('test', node=node) as c:
            try:
                c.assert_true(c.get_value('.not.a.path'))
            except NoDataError:
                pass

            c.assert_true(True)

        captured = capsys.readouterr()
        result = json.loads(captured.out)
        assert len(result['assertions']) == 1

    def test_suppress_on_no_data_get(self, capsys):
        node = Node.from_component_json({})

        with Check('test', node=node) as c:
            try:
                c.get_value('.not.a.path')
            except NoDataError:
                pass

            c.assert_true(True)

        captured = capsys.readouterr()
        result = json.loads(captured.out)
        assert len(result['assertions']) == 1

    def test_no_data_on_exists_assertion_workflows_not_finished(self, capsys):
        node = Node.from_component_json({}, bundle_info={'workflows_finished': False})

        with Check('test', node=node) as c:
            c.assert_exists('.not.a.path')
            c.assert_equals('should not run', 'should not run')

        captured = capsys.readouterr()
        result = json.loads(captured.out)
        assert len(result['assertions']) == 1
        assert result['assertions'][0]['result'] == 'no-data'

    def test_no_data_on_exists_assertion_workflows_finished(self, capsys):
        node = Node.from_component_json({}, bundle_info={'workflows_finished': True})

        with Check('test', node=node) as c:
            c.assert_exists('.not.a.path')
            c.assert_equals('should run', 'should run')

        captured = capsys.readouterr()
        result = json.loads(captured.out)
        assert len(result['assertions']) == 2
        assert result['assertions'][0]['result'] == 'fail'



import importlib.metadata
from .node import Node
from .nodepath import NodePath
from .check import Check
from .result import Op, CheckStatus, AssertionResult, NoDataError
from .variables import variable, variable_or_default

__version__ = importlib.metadata.version('lunar_policy')

__all__ = [
    # Nodes
    'Node',
    'NodePath',
    # Making Assertions
    'Check',
    # Result Types
    'Op',
    'CheckStatus',
    'AssertionResult',
    # Snippet Variables
    'variable',
    'variable_or_default',
    # Exceptions
    'NoDataError',
]



from enum import Enum
from dataclasses import dataclass
from typing import Any, Optional
from dataclasses import asdict


class NoDataError(Exception):
    pass


class CheckStatus(Enum):
    PASS = 'pass'
    FAIL = 'fail'
    PENDING = 'no-data'
    ERROR = 'error'


class Op(Enum):
    CONTAINS = 'contains'
    EQUALS = 'equals'
    TRUE = 'true'
    FALSE = 'false'
    GREATER = 'greater'
    GREATER_OR_EQUAL = 'greater_or_equal'
    LESS = 'less'
    LESS_OR_EQUAL = 'less_or_equal'
    MATCH = 'match'
    FAIL = 'fail'
    ERROR = 'error'
    EXISTS = 'exists'


@dataclass
class AssertionResult:
    op: Op
    args: list[Any]
    result: CheckStatus
    failure_message: Optional[str] = None

    def toJson(self):
        result = asdict(self)

        result['op'] = self.op.value
        result['result'] = self.result.value

        if result['failure_message'] is None:
            del result['failure_message']

        return result



import json
import pytest

from src.lunar_policy import Check, Node


class TestCheckBasic:
    def test_invalid_check_initialization(self):
        with pytest.raises(ValueError):
            Check('test', node='not a SnippetData object')

    def test_data_error_if_no_env(
        self,
    ):
        with pytest.raises(ValueError):
            Check('test', node=None)

    def test_data_error_if_invalid_path(self, monkeypatch):
        monkeypatch.setenv('LUNAR_BUNDLE_PATH', '/invalid/path')

        with pytest.raises(ValueError):
            Check('test')

    def test_data_is_set(self, capsys):
        dataJson = {'hi': 'there'}
        node = Node.from_component_json(dataJson)

        with Check('test', node=node) as c:
            c.get_value('.hi')

    def test_description_check(self, capsys):
        node = Node.from_component_json({})

        with Check('test', 'description', node=node) as c:
            c.assert_true(True)

        captured = capsys.readouterr()
        result = json.loads(captured.out)

        assert result['description'] == 'description'

    def test_description_not_in_check(self, capsys):
        node = Node.from_component_json({})

        with Check('test', node=node) as c:
            c.assert_true(True)

        captured = capsys.readouterr()
        result = json.loads(captured.out)

        assert 'description' not in result



import json
from typing import Optional, Any, List
from pathlib import Path as FSPath
from .nodepath import NodePath
from .result import NoDataError


class NodePathTracker:
    def __init__(self):
        self.path_tracker = []

    def track(self, path: NodePath):
        self.path_tracker.append(str(path))

    def get_paths(self) -> List[NodePath]:
        return self.path_tracker


class Node:
    # Root path constant to avoid repeated NodePath.parse(".") calls
    _ROOT_PATH = NodePath.parse('.')

    def __init__(self, path: NodePath, data: dict, path_tracker: NodePathTracker):
        if not data:
            raise ValueError('data dictionary cannot be empty')
        if 'merged_blob' not in data:
            raise ValueError("data must contain 'merged_blob' key")
        if 'metadata_instances' not in data:
            raise ValueError("data must contain 'metadata_instances' key")
        if 'bundle_info' not in data:
            raise ValueError("data must contain 'bundle_info' key")

        if not path_tracker:
            raise ValueError('path_tracker cannot be None')

        self._path = path
        self._data = data
        self._path_tracker = path_tracker

    @classmethod
    def from_bundle_file(
        cls,
        file_path: FSPath,
    ) -> 'Node':
        if not file_path.is_absolute():
            raise ValueError('path must be an absolute path')

        with file_path.open() as f:
            data = json.load(f)

        return cls(cls._ROOT_PATH, data, NodePathTracker())

    @classmethod
    def from_bundle_json(
        cls,
        data: dict,
    ) -> 'Node':
        if not isinstance(data, dict):
            raise ValueError('data must be a dictionary')

        return cls(cls._ROOT_PATH, data, NodePathTracker())

    @classmethod
    def from_component_json_file(
        cls,
        file_path: FSPath,
        bundle_info: Optional[dict] = None,
    ) -> 'Node':
        with file_path.open() as f:
            data = json.load(f)

        if bundle_info is None:
            bundle_info = {}

        return cls(
            cls._ROOT_PATH,
            {
                'bundle_info': bundle_info,
                'merged_blob': data,
                'metadata_instances': [{'payload': data}],
            },
            NodePathTracker(),
        )

    @classmethod
    def from_component_json(
        cls,
        data: dict,
        bundle_info: Optional[dict] = None,
    ) -> 'Node':
        if bundle_info is None:
            bundle_info = {}

        if not isinstance(data, dict):
            raise ValueError('data must be a dictionary')

        return cls(
            cls._ROOT_PATH,
            {
                'bundle_info': bundle_info,
                'merged_blob': data,
                'metadata_instances': [{'payload': data}],
            },
            NodePathTracker(),
        )

    def _workflows_finished(self) -> bool:
        bundle_info = self._data.get('bundle_info', {})
        if not bundle_info:
            return False

        return bundle_info.get('workflows_finished', False)

    def get_node(self, path: str) -> 'Node':
        new_path = self._path.concat(path)
        return self.__class__(new_path, self._data, self._path_tracker)

    def exists(self, path: str = '.') -> bool:
        try:
            self.get_value(path)
            return True
        except ValueError:
            return False

    def get_value(self, path: str = '.') -> Any:
        full_path = self._path.concat(path)
        self._path_tracker.track(full_path)

        result = full_path.get_value(self._data['merged_blob'])

        if result is None:
            if self._workflows_finished():
                raise ValueError(f'No data found at path {path}')
            else:
                raise NoDataError(f'No data found at path {path}')

        return result

    def get_all_values(self, query: str = '.') -> List[Any]:
        full_path = self._path.concat(query)
        self._path_tracker.track(full_path)

        results = []
        for instance in reversed(self._data['metadata_instances']):
            payload = instance.get('payload', {})
            value = full_path.get_value(payload)

            if value is None:
                continue

            results.append(value)

        if len(results) == 0:
            if self._workflows_finished():
                raise ValueError(f'No data found for {query}')
            else:
                raise NoDataError(f'No data found for {query}')

        return results

    def __iter__(self):
        current_value = self.get_value()

        if isinstance(current_value, list):
            for i, item in enumerate(current_value):
                yield self.get_node(f'[{i}]')
            return

        if isinstance(current_value, dict):
            for key in current_value.keys():
                yield key
            return

        raise ValueError(f'Cannot iterate over scalar value: {type(current_value).__name__}')

    def items(self):
        current_value = self.get_value()

        if not isinstance(current_value, dict):
            raise ValueError('items() is only valid for dict-like data')

        for key in current_value.keys():
            yield (key, self.get_node(f'.{key}'))



import json
import pytest

from src.lunar_policy import Check, Node


class TestCheckGet:
    def test_get_valid_path_workflows_finished(self):
        node = Node.from_component_json({'hi': 'there'}, bundle_info={'workflows_finished': True})

        with Check('test', node=node) as c:
            c.get_value('.hi')

    def test_get_valid_path_workflows_not_finished(self):
        node = Node.from_component_json({'hi': 'there'}, bundle_info={'workflows_finished': False})

        with Check('test', node=node) as c:
            c.get_value('.hi')

    def test_get_missing_path_workflows_finished(self):
        node = Node.from_component_json({}, bundle_info={'workflows_finished': True})

        with pytest.raises(ValueError):
            with Check('test', node=node) as c:
                c.get_value('.missing')

    def test_get_missing_path_workflows_not_finished(self, capsys):
        node = Node.from_component_json({}, bundle_info={'workflows_finished': False})

        with Check('test', node=node) as c:
            c.get_value('.missing')

        captured = capsys.readouterr()
        results = json.loads(captured.out)['assertions']
        assert len(results) == 1
        assert results[0]['result'] == 'no-data'
        assert '.missing' in results[0]['failure_message']
        assert results[0]['op'] == 'fail'

    def test_get_invalid_path(self):
        node = Node.from_component_json({})

        with pytest.raises(ValueError):
            with Check('test', node=node) as c:
                c.get_value('@#$@#$@#$')

    def test_get_all_valid_path_single(self):
        node = Node.from_bundle_json(
            {
                'bundle_info': {},
                'merged_blob': {},
                'metadata_instances': [{'payload': {'hi': 'there'}}],
            }
        )

        all = []
        with Check('test', node=node) as c:
            all = c.get_all_values('.hi')

        assert all == ['there']

    def test_get_all_valid_path_multiple(self):
        node = Node.from_bundle_json(
            {
                'bundle_info': {},
                'merged_blob': {},
                'metadata_instances': [
                    {'payload': {'hi': 'there'}},
                    {'payload': {'hi': 'there'}},
                ],
            }
        )

        all = []
        with Check('test', node=node) as c:
            all = c.get_all_values('.hi')

        assert all == ['there', 'there']

    def test_get_all_invalid_path(self):
        node = Node.from_component_json({})

        with pytest.raises(ValueError):
            with Check('test', node=node) as c:
                c.get_all_values('@#$@#$@#$')


class TestCheckIteration:
    def test_iteration(self):
        node = Node.from_component_json({'hi': 'there'})

        with Check('test', node=node) as c:
            for key in c:
                assert key == 'hi'

    def test_items(self):
        node = Node.from_component_json({'hi': 'there'})

        with Check('test', node=node) as c:
            for key, node in c.items():
                assert key == 'hi'
                assert node.get_value() == 'there'
